#define _FILE_OFFSET_BITS 64

#include "Parameters.h"
#include "Pedigree.h"
#include "Error.h"
#include "Random.h"
#include "MathVector.h"
#include "ShotGun.h"
#include "gsl/gsl_sf_gamma.h"
#include "gsl/gsl_rng.h"
#include "gsl/gsl_randist.h"
#include "config.h"

int main(int argc, char ** argv)
{
setbuf(stdout, NULL);
setbuf(stderr, NULL);
   String hapfile, posfile, outfile("sgReads");
   double beginpos, endpos, times = 1.0, erate = 0.005, complete = 1.00, alpha = 0.001;
   int readlength = 32, seed = 12345, n = 10000, B = 5000, minwts = 2;
   int numberReads;
   bool allmarkers = false, truth = false;

   printf("Shotgun Sequencing from Cosi 0.0.6\n"
          "2 = ancestral allele; 1 = new allele\n"
          "            (c) 2007 - 2010 Yun Li\n");

   ParameterList pl;

BEGIN_LONG_PARAMETERS(longParameters)
   LONG_PARAMETER_GROUP("Input Files")
      LONG_STRINGPARAMETER("hapfile", &hapfile)
      LONG_STRINGPARAMETER("posfile", &posfile)
   LONG_PARAMETER_GROUP("Options")
      LONG_INTPARAMETER("seed", &seed)
      LONG_INTPARAMETER("readlength", &readlength)
      LONG_INTPARAMETER("minwts", &minwts)
      LONG_DOUBLEPARAMETER("BeginningPosition", &beginpos)
      LONG_DOUBLEPARAMETER("EndingPosition", &endpos)
      LONG_DOUBLEPARAMETER("complete", &complete)
      LONG_DOUBLEPARAMETER("alpha", &alpha)
      LONG_INTPARAMETER("numIndividuals[n]", &n)
      LONG_DOUBLEPARAMETER("times[t]", &times)
      LONG_DOUBLEPARAMETER("erate", &erate)
      LONG_PARAMETER("all", &allmarkers)
      LONG_PARAMETER("trueped", &truth)
      LONG_INTPARAMETER("numReplications",&B)
   LONG_PARAMETER_GROUP("Output")
      LONG_STRINGPARAMETER("prefix", &outfile)
END_LONG_PARAMETERS();

   pl.Add(new LongParameters("Available Options", longParameters));

   pl.Add(new HiddenString('h', "Haplotype File", hapfile));
   pl.Add(new HiddenString('p', "Position File", posfile));
   pl.Add(new HiddenInteger('s', "Seed", seed));
   pl.Add(new HiddenInteger('n', "numIndividuals", n));
   pl.Add(new HiddenDouble('t', "Times", times));
   pl.Add(new HiddenDouble('b', "Beginning Position", beginpos));
   pl.Add(new HiddenDouble('e', "Ending Position", endpos));
   pl.Add(new HiddenString('o', "Output prefix", outfile));

   pl.Read(argc, argv);
   pl.Status();

   if ( (int) (B * alpha ) < 5) printf("WARNING: consider increasing numReplications\n");
   
   globalRandom.Reset(seed);

  // pick score cutoff to control false positive rate
  gsl_rng *r = gsl_rng_alloc(gsl_rng_mt19937);
  gsl_rng_set(r, seed);
  Vector scores(B);
  //float * scores = new float [B];
  for (int i = 0; i < B; i ++)
  {
    //if (i % 1000 == 0) printf("%d\n", i);
    scores[i] = 0;
    for (int j = 0; j < n; j++)
    {
      int rpois = gsl_ran_poisson(r, times * erate);
      scores[i] += ( rpois *(rpois + 1) / 2.0);
    }
  }
  scores.Sort();
  scores.Reverse();
  int CutoffIndex = (int) (B * alpha);
  minwts = (int) scores[CutoffIndex];
  scores.Delete(B);

  //printf("AA: %d\n", minwts);
  gsl_rng_free(r);

   // Load haplotypes, positions, 
   ShotGun sg;
   sg.readlength = readlength;
   
   sg.LoadMarkerPos(posfile);
   sg.LoadHaplotypes(hapfile,n);
   
   numberReads = (int)  ((endpos - beginpos) * times) / readlength;
   if (sg.count != 0 && sg.markerCount != 0)
     printf("%2.1fX sequencing for %d individuals across %d markers (%d reads)\n", 
            times, sg.count / 2, sg.markerCount, numberReads);
   
   sg.ShotGunRead(beginpos, endpos, numberReads, erate, complete);
   
   if (sg.markerCount == 0 || sg.count == 0)
   {
	   printf("\nNo valid data\n");
	   return 0;
   }
   
   sg.PassHit(minwts);
   
   FILE * file;
   
   if (allmarkers)
   {
	   file = fopen(outfile + ".all.cnts", "wt");
	   if (file == NULL)
	     error("Error opening output file [%s.all.cnts]\n", (const char *) outfile);
	   else
	   {
		   for (int i = 0; i < sg.count / 2; i++)
		   {
			   int indivID = i + 1;
			   fprintf(file, "%d %d 0 0 1 ", indivID, indivID);
			   for (int j = 0; j < sg.bp; j++)
			   {
			     fprintf(file, "%d %d ", sg.alleleOneCounts[i][j], sg.alleleTwoCounts[i][j]);
			   }
			   fprintf(file, "\n");
		   }
	   }
	   fclose(file);
	   
	   file = fopen(outfile + ".all.info", "wt");
	   if (file == NULL)
	     error("Error opening output file [%s.all.info]\n", (const char *) outfile);
	   else
	   {
		   //fprintf(file, "SNP passHit Al1 Al2\n");
		   for (int i = 0; i < sg.bp; i++)
		   {
			   fprintf(file, "M%d %d 1 2\n", i, sg.isPassHit[i]);
		   }
	   }
	   fclose(file);
	
   }
   
   if (sg.passHitMarkerCount == 0)
   {
	   printf("\nNo >=pass-hit markers\n");
	   return 1;
   }
   
   file = fopen(outfile + ".phits.cnts", "wt");
   if (file == NULL)
     error("Error opening output file [%s.phits.cnts]\n", (const char *) outfile);
   else
   {
	   for (int i = 0; i < sg.count / 2; i++)
	   {
		   int indivID = i + 1;
		   fprintf(file, "%d %d 0 0 1 ", indivID, indivID);
		   for (int j = 0; j < sg.bp; j++)
		   {
			   if (sg.isPassHit[j])
			     fprintf(file, "%d %d ", sg.alleleOneCounts[i][j], sg.alleleTwoCounts[i][j]);
		   }
		   fprintf(file, "\n");
	   }
   }
   fclose(file);
   
   file = fopen(outfile + ".phits.info", "wt");
   if (file == NULL)
     error("Error opening output file [%s.phits.info]\n", (const char *) outfile);
   else
   {
	   //fprintf(file, "SNP Al1 Al2\n");
	   for (int i = 0; i < sg.bp; i++)
	   {
		   if (sg.isPassHit[i])
		   {
			   fprintf(file, "M%d 1 2\n", i);
		   }
		     
	   }
   }
   fclose(file);
   
   if (truth)
   {
	   file = fopen(outfile + ".true_phits.ped", "wt");
	   if (file == NULL)
	     error("Error opening output file [%s.true_phits.cnts]\n", (const char *) outfile);
	   else
	   {
		   for (int i = 0; i < sg.count / 2; i++)
		   {
			   int indivID = i + 1;
			   fprintf(file, "%d %d 0 0 1 ", indivID, indivID);
			   for (int j = 0; j < sg.bp; j++)
			   {
				   if (sg.isPassHit[j])
				     fprintf(file, "%d/%d ", sg.sequencesOne[i][j], sg.sequencesTwo[i][j]);
			   }
			   fprintf(file, "\n");
		   }
	   }
	   fclose(file);
   }
}
