#include <stdint.h>
#include "ShotGun.h"
#include "MemoryAllocators.h"
#include "MemoryInfo.h"
#include "Error.h"
#include "InputFile.h"

#ifndef ZEPS
#define ZEPS 1e-30
#endif

ShotGun::ShotGun()
   {
   haplotypes = NULL;
   bp = 0;
   markerCount = 0;
   count = 0;
   alleleOneCounts = NULL;
   alleleTwoCounts = NULL;
   isPassHit = NULL;
   passHitMarkerCount = 0;
   sequencesOne = NULL;
   sequencesTwo = NULL;
   }

ShotGun::~ShotGun()
   {
   if (haplotypes != NULL)
      FreeCharMatrix(haplotypes, count + 1);
   if (alleleOneCounts != NULL)
      FreeCharMatrix(alleleOneCounts, count/2);
   if (alleleTwoCounts != NULL)
      FreeCharMatrix(alleleTwoCounts, count/2);
   if (sequencesOne != NULL)
      FreeCharMatrix(sequencesOne, count/2);
   if (sequencesTwo != NULL)
      FreeCharMatrix(sequencesTwo, count/2);
   if (isPassHit != NULL)
      delete [] isPassHit;
   }

void ShotGun::LoadMarkerPos(const char * filename)
   {
   IFILE f = ifopen(filename, "rb");

   if (f == NULL)
      return; // error("Marker list [%s] could not be opened\n", filename);

   LoadMarkerPos(f);
   ifclose(f);
   }

void ShotGun::LoadHaplotypes(const char * filename, int numIndiv)
   {
   IFILE f = ifopen(filename, "rb");

   if (f == NULL)
      {
      if (Pedigree::markerCount)
         error("File [%s] with phased haplotypes could not be opened\n", filename);

      return;
      }

   LoadHaplotypes(f, numIndiv);
   ifclose(f);
   }

void ShotGun::LoadMarkerPos(IFILE file)
   {
   String      buffer;
   StringArray tokens;

   printf("Loading list of markers in phased haplotype ...\n");

   while (!ifeof(file))
      {
      buffer.ReadLine(file);
      tokens.ReplaceTokens(buffer);

      if (tokens.Length() == 0) continue;

      if (tokens.Length() > 2)
         error("Each line should list exactly two fields: marker name, and \n"
               "physical position (in bp), but the following line appears \n"
               "to include extra information:\n\n"
               "%s\n", (const char *) buffer);

      int markerId = Pedigree::GetMarkerID(tokens[0]);

      if (markerCount++ != markerId)
         error("Marker %s is duplicated.\n\n"
               "Every marker should have a unique name.\n",
               (const char *) tokens[0]);
               
      MarkerInfo * info = Pedigree::GetMarkerInfo(tokens[0]);
      info->position = (double) tokens[1];
      }
   }

void ShotGun::LoadHaplotypes(IFILE file, int numIndiv)
   {
   // Don't load haplotypes unless we have a marker list
   if (markerCount == 0)
      {
      printf("  WARNING -- Since no marker list was provided, haplotype file will be ignored\n\n");
      return;
      }

   printf("Loading phased haplotypes ...\n");

   String      buffer;
   StringArray tokens;

   // In the first pass, we simply count the number of non-blank lines
   while (!ifeof(file) && count < (numIndiv * 2) )
      {
      buffer.ReadLine(file);
      tokens.ReplaceTokens(buffer);

      if (!tokens.Length()) continue;

      count++;
      }

   // Check if we got some valid input
   if (count == 0 || markerCount == 0)
      return;

   // Then, we allocate memory for storing the phased haplotypes
   haplotypes = AllocateCharMatrix(count + 1, Pedigree::markerCount);

   // And finally, we load the data in a second pass
   ifrewind(file);

   int line = 0, index = 0;
   while (!ifeof(file) && line < (numIndiv * 2) )
      {
      line++;
      buffer.ReadLine(file);
      tokens.ReplaceTokens(buffer);

      if (tokens.Length() == 0) continue;

      int hapstart = tokens.Length() - 1;
      int offset = markerCount;

      while ((offset -= tokens[hapstart].Length()) > 0 && hapstart > 0)
         hapstart--;

      if (offset != 0)
         error("The haplotype file format was not recognized\n"
               "(Problem occured reading haplotype #%d in line #%d)\n\n"
               "Check that the number of markers matches the SNPs list\n",
               ++line, ++index);

      for (int i = 0; i < markerCount; i++)
         {
         MarkerInfo * info = Pedigree::GetMarkerInfo(i);

         if (offset == tokens[hapstart].Length())
            {
            offset = 0;
            hapstart++;
            }

         int al = info->GetAlleleNumber(tokens[hapstart][offset++]);

         if (al == -1)
            al = info->NewAllele(tokens[hapstart][offset-1]);

         if (al == 0)
            error("Missing data in haplotype %d at position %d (marker %s)\n",
                   ++index, ++i, (const char *) Pedigree::markerNames[i]);

         if (al > 2)
            error("More than 2 alleles at position %d (marker %s)\n",
                   ++i, (const char *) Pedigree::markerNames[i]);

         haplotypes[index][i] = al;
         }
      index++;
      }

   for (int i = 0; i < markerCount; i++)
      {
      MarkerInfo * info = Pedigree::GetMarkerInfo(i);

      // The last row in the haplotypes file stores the number of distinct
      // alleles observed in the input files
      haplotypes[count][i] = info->CountAlleles();
      }
      
   }

void ShotGun::ShotGunRead(double start, double end, int reads, double erate, double complete, Random * rand)
{
	bp = (int) end - (int) start + 1;
	
	if (bp == 0)
	  error("Total Length = ZERO bp??\n");
	
	if (haplotypes != NULL)
	{
		sequencesOne = AllocateCharMatrix(count / 2, bp);
		sequencesTwo = AllocateCharMatrix(count / 2, bp);
		
		if (sequencesOne == NULL || sequencesTwo == NULL)
		  error("Failed to allocate memory for sequences\n");
		  
		// initialize: ancestral allele = 2 
		for (int i = 0; i < count / 2; i++)
		{
			int chrOne = 2 * i;
			int chrTwo = chrOne + 1;
			for (int j = 0; j < bp; j++)
			{
				sequencesOne[i][j] = sequencesTwo[i][j] = 2;
			}
			// update according to loaded cosi haplotypes
			for (int k = 0; k < markerCount; k++)
			{
				MarkerInfo * info = Pedigree::GetMarkerInfo(k);
				int bpIndex = (int)info->position - (int) start;
				if (bpIndex >= bp)
				  // error("marker %s at position %d out of range?", (const char *) info->name, (int)info->position);
				  continue;

				if ( (int) info->GetAlleleLabel((int) haplotypes[chrOne][k]) == 1)
				  sequencesOne[i][bpIndex] = 1;
				if ( (int) info->GetAlleleLabel((int) haplotypes[chrTwo][k]) == 1)
				  sequencesTwo[i][bpIndex] = 1;
			}
		}
	}
	
   if (haplotypes != NULL)
      FreeCharMatrix(haplotypes, count + 1);


	// initialize 
	alleleOneCounts = AllocateCharMatrix(count/2, bp);
	alleleTwoCounts = AllocateCharMatrix(count/2, bp);
	if (alleleOneCounts == NULL || alleleTwoCounts == NULL)
	  error("Failed to allocate memory for counts\n");
	for (int i = 0; i < count/2 ; i++)
		for (int j = 0; j < bp; j++)
			alleleOneCounts[i][j] = alleleTwoCounts[i][j] = 0;
			

	// update
	if (rand == NULL)
		rand = &globalRandom;
		
	for (int i = 0; i < count/2 ; i++)
	{
		for (int j = 0; j < reads; j++)
		{
			int startingPt = rand->NextInt() % bp;
			// randomly pick one chromosome to read
			int chr = rand->NextInt() % 2;
			char ** sequences = sequencesOne;
			if (chr == 1)
			  sequences = sequencesTwo;
			  
			for (int k = startingPt; k < (startingPt + readlength); k++)
			{
				if ( k >= bp)
				  continue;
				
				if (sequences[i][k] == 2)
				{
					if(rand->Uniform() > erate)
					  alleleTwoCounts[i][k]++;
					else
					  alleleOneCounts[i][k]++;
				}
				else
				{
					if(rand->Uniform() < erate)
					  alleleTwoCounts[i][k]++;
					else
					  alleleOneCounts[i][k]++;
				}
			}//end of k
		}//end if j
		//printf("\n");
		printf("Individual %d/%d\r", i, count/2);
		fflush(stdout);
	}//end of i
	
	if (complete < 1.0)
	{
  	for (int j = 0; j < bp; j++)
  	{
    	if ( rand->Uniform() > complete)
    	{
      	for (int i = 0; i < count/2 ; i++)
      	{
        	alleleOneCounts[i][j] = alleleTwoCounts[i][j] = 0;
      	}
    	}
  	}
	}
}


//nwts <= 255
void ShotGun::PassHit(int nwts)
{
	if (nwts > 65535)
	   error("minwts > 255 NOT allowed\n\n");
	   
	if (bp)
	{
		isPassHit = new char [bp];
		uint16_t * sumAlleleOneWts = new uint16_t [bp];
		uint16_t * sumAlleleTwoWts = new uint16_t [bp];
		
		if (isPassHit == NULL || sumAlleleOneWts == NULL || sumAlleleTwoWts == NULL )
		  error("Failed to allocate memory for hits\n");
		  
		else
		{
			// initialize
			for (int i = 0; i < bp; i++)
			{
				isPassHit[i] = 0;
				sumAlleleOneWts[i] = 0;
				sumAlleleTwoWts[i] = 0;
			}
			
			// search through individuals
			for (int i = 0; i < bp; i++)
			{
				for (int j = 0; j < count/2; j++)
				{
					if (isPassHit[i])
					  break;
					  
					if ( sumAlleleOneWts[i] < nwts) 
					for (int k = 1; k <= alleleOneCounts[j][i] ; k++) 
					{
						sumAlleleOneWts[i] += k;
					}
					if (sumAlleleTwoWts[i] < nwts) 
					for (int k = 1; k <= alleleTwoCounts[j][i] ; k++) 
					{
						sumAlleleTwoWts[i] += k;
					}
					
					// condition for passing:
					//   total minor weights >= nwts  
					if ( (sumAlleleOneWts[i] >= nwts && sumAlleleTwoWts[i] >= nwts))
					{
						isPassHit[i] = 1;
						passHitMarkerCount ++;
					}
				}
			}
			
			if (sumAlleleOneWts != NULL) 
			  delete [] sumAlleleOneWts;
			if (sumAlleleTwoWts != NULL) 
			  delete [] sumAlleleTwoWts;
		}
	}
}


